"""
ANSI -> html converter

Credit for original idea and implementation
goes to Muhammad Alkarouri and his
snippet #577349 on http://code.activestate.com.

(extensively modified by Griatch 2010)
"""

import re
from html import escape as html_escape

from .ansi import *

# All xterm256 RGB equivalents

XTERM256_FG = "\033[38;5;{}m"
XTERM256_BG = "\033[48;5;{}m"


class TextToHTMLparser(object):
    """
    This class describes a parser for converting from ANSI to html.
    """

    tabstop = 4

    style_codes = [
        # non-color style markers
        ANSI_NORMAL,
        ANSI_UNDERLINE,
        ANSI_HILITE,
        ANSI_UNHILITE,
        ANSI_INVERSE,
        ANSI_BLINK,
        ANSI_INV_HILITE,
        ANSI_BLINK_HILITE,
        ANSI_INV_BLINK,
        ANSI_INV_BLINK_HILITE,
    ]

    ansi_color_codes = [
        # Foreground colors
        ANSI_BLACK,
        ANSI_RED,
        ANSI_GREEN,
        ANSI_YELLOW,
        ANSI_BLUE,
        ANSI_MAGENTA,
        ANSI_CYAN,
        ANSI_WHITE,
    ]

    xterm_fg_codes = [XTERM256_FG.format(i + 16) for i in range(240)]

    ansi_bg_codes = [
        # Background colors
        ANSI_BACK_BLACK,
        ANSI_BACK_RED,
        ANSI_BACK_GREEN,
        ANSI_BACK_YELLOW,
        ANSI_BACK_BLUE,
        ANSI_BACK_MAGENTA,
        ANSI_BACK_CYAN,
        ANSI_BACK_WHITE,
    ]

    xterm_bg_codes = [XTERM256_BG.format(i + 16) for i in range(240)]

    re_style = re.compile(
        r"({})".format(
            "|".join(
                style_codes + ansi_color_codes + xterm_fg_codes + ansi_bg_codes + xterm_bg_codes
            ).replace("[", r"\[")
        )
    )

    colorlist = (
        [ANSI_UNHILITE + code for code in ansi_color_codes]
        + [ANSI_HILITE + code for code in ansi_color_codes]
        + xterm_fg_codes
    )

    bglist = ansi_bg_codes + [ANSI_HILITE + code for code in ansi_bg_codes] + xterm_bg_codes

    re_string = re.compile(
        r"(?P<htmlchars>[<&>])|(?P<tab>[\t]+)|(?P<lineend>\r\n|\r|\n)",
        re.S | re.M | re.I,
    )
    re_url = re.compile(
        r'(?<!=")(\b(?:ftp|www|https?)\W+(?:(?!\.(?:\s|$)|&\w+;)[^"\',;$*^\\(){}<>\[\]\s])+)(\.(?:\s|$)|&\w+;|)'
    )
    re_protocol = re.compile(r"^(?:ftp|https?)://")
    re_valid_no_protocol = re.compile(
        r"^(?:www|ftp)\.[-a-zA-Z0-9@:%._\+~#=]{2,256}\.[a-z]{2,6}\b[-a-zA-Z0-9@:%_\+.~#?&//=]*"
    )
    re_mxplink = re.compile(r"\|lc(.*?)\|lt(.*?)\|le", re.DOTALL)
    re_mxpurl = re.compile(r"\|lu(.*?)\|lt(.*?)\|le", re.DOTALL)

    def remove_bells(self, text):
        """
        Remove ansi specials

        Args:
            text (str): Text to process.

        Returns:
            text (str): Processed text.

        """
        return text.replace(ANSI_BEEP, "")

    def remove_backspaces(self, text):
        """
        Removes special escape sequences

        Args:
            text (str): Text to process.

        Returns:
            text (str): Processed text.

        """
        backspace_or_eol = r"(.\010)|(\033\[K)"
        n = 1
        while n > 0:
            text, n = re.subn(backspace_or_eol, "", text, 1)
        return text

    def convert_linebreaks(self, text):
        """
        Extra method for cleaning linebreaks

        Args:
            text (str): Text to process.

        Returns:
            text (str): Processed text.

        """
        return text.replace("\n", r"<br>")

    def convert_urls(self, text):
        """
        Replace urls (http://...) by valid HTML.

        Args:
            text (str): Text to process.

        Returns:
            text (str): Processed text.

        """
        m = self.re_url.search(text)
        if m:
            href = m.group(1)
            label = href
            # if there is no protocol (i.e. starts with www or ftp)
            # prefix with http:// so the link isn't treated as relative
            if not self.re_protocol.match(href):
                if not self.re_valid_no_protocol.match(href):
                    return text
                href = "http://" + href
            rest = m.group(2)
            # -> added target to output prevent the web browser from attempting to
            # change pages (and losing our webclient session).
            return (
                text[: m.start()]
                + f'<a href="{href}" target="_blank">{label}</a>{rest}'
                + text[m.end() :]
            )
        else:
            return text

    def sub_mxp_links(self, match):
        """
        Helper method to be passed to re.sub,
        replaces MXP links with HTML code.

        Args:
            match (re.Matchobject): Match for substitution.

        Returns:
            text (str): Processed text.

        """
        cmd, text = [grp.replace('"', "\\&quot;") for grp in match.groups()]
        val = (
            r"""<a id="mxplink" href="#" """
            """onclick="Evennia.msg(&quot;text&quot;,[&quot;{cmd}&quot;],{{}});"""
            """return false;">{text}</a>""".format(cmd=cmd, text=text)
        )
        return val

    def sub_mxp_urls(self, match):
        """
        Helper method to be passed to re.sub,
        replaces MXP links with HTML code.
        Args:
            match (re.Matchobject): Match for substitution.
        Returns:
            text (str): Processed text.
        """
        url, text = [grp.replace('"', "\\&quot;") for grp in match.groups()]
        val = r"""<a id="mxplink" href="{url}" target="_blank">{text}</a>""".format(
            url=url, text=text
        )
        return val

    def sub_text(self, match):
        """
        Helper method to be passed to re.sub,
        for handling all substitutions.

        Args:
            match (re.Matchobject): Match for substitution.

        Returns:
            text (str): Processed text.

        """
        cdict = match.groupdict()
        if cdict["htmlchars"]:
            return html_escape(cdict["htmlchars"])
        elif cdict["lineend"]:
            return "<br>"
        elif cdict["tab"]:
            text = cdict["tab"].replace("\t", " " * (self.tabstop))
            return text
        return None

    def format_styles(self, text):
        """
        Takes a string with parsed ANSI codes and replaces them with
        HTML spans and CSS classes.

        Args:
            text (str): The string to process.

        Returns:
            text (str): Processed text.
        """

        # split out the ANSI codes and clean out any empty items
        str_list = [substr for substr in self.re_style.split(text) if substr]
        # initialize all the flags and classes
        classes = []
        clean = True
        inverse = False
        # default color is light grey - unhilite + white
        hilight = ANSI_UNHILITE
        fg = ANSI_WHITE
        # default bg is black
        bg = ANSI_BACK_BLACK

        for i, substr in enumerate(str_list):
            # reset all current styling
            if substr == ANSI_NORMAL:
                # close any existing span if necessary
                str_list[i] = "</span>" if not clean else ""
                # reset to defaults
                classes = []
                clean = True
                inverse = False
                hilight = ANSI_UNHILITE
                fg = ANSI_WHITE
                bg = ANSI_BACK_BLACK

            # change color
            elif substr in self.ansi_color_codes + self.xterm_fg_codes:
                # erase ANSI code from output
                str_list[i] = ""
                # set new color
                fg = substr

            # change bg color
            elif substr in self.ansi_bg_codes + self.xterm_bg_codes:
                # erase ANSI code from output
                str_list[i] = ""
                # set new bg
                bg = substr

            # non-color codes
            elif substr in self.style_codes:
                # erase ANSI code from output
                str_list[i] = ""

                # hilight codes
                if substr in (ANSI_HILITE, ANSI_UNHILITE, ANSI_INV_HILITE, ANSI_INV_BLINK_HILITE):
                    # set new hilight status
                    hilight = ANSI_UNHILITE if substr == ANSI_UNHILITE else ANSI_HILITE

                # inversion codes
                if substr in (ANSI_INVERSE, ANSI_INV_HILITE, ANSI_INV_BLINK_HILITE):
                    inverse = True

                # blink codes
                if (
                    substr in (ANSI_BLINK, ANSI_BLINK_HILITE, ANSI_INV_BLINK_HILITE)
                    and "blink" not in classes
                ):
                    classes.append("blink")

                # underline
                if substr == ANSI_UNDERLINE and "underline" not in classes:
                    classes.append("underline")

            else:
                # normal text, add text back to list
                if not str_list[i - 1]:
                    # prior entry was cleared, which means style change
                    # get indices for the fg and bg codes
                    bg_index = self.bglist.index(bg)
                    try:
                        color_index = self.colorlist.index(hilight + fg)
                    except ValueError:
                        # xterm256 colors don't have the hilight codes
                        color_index = self.colorlist.index(fg)

                    if inverse:
                        # inverse means swap fg and bg indices
                        bg_class = "bgcolor-{}".format(str(color_index).rjust(3, "0"))
                        color_class = "color-{}".format(str(bg_index).rjust(3, "0"))
                    else:
                        # use fg and bg indices for classes
                        bg_class = "bgcolor-{}".format(str(bg_index).rjust(3, "0"))
                        color_class = "color-{}".format(str(color_index).rjust(3, "0"))

                    # black bg is the default, don't explicitly style
                    if bg_class != "bgcolor-000":
                        classes.append(bg_class)
                    # light grey text is the default, don't explicitly style
                    if color_class != "color-007":
                        classes.append(color_class)
                    # define the new style span
                    prefix = '<span class="{}">'.format(" ".join(classes))
                    # close any prior span
                    if not clean:
                        prefix = "</span>" + prefix
                    # add span to output
                    str_list[i - 1] = prefix

                    # clean out color classes to easily update next time
                    classes = [cls for cls in classes if "color" not in cls]
                    # flag as currently being styled
                    clean = False

        # close span if necessary
        if not clean:
            str_list.append("</span>")
        # recombine back into string
        return "".join(str_list)

    def parse(self, text, strip_ansi=False):
        """
        Main access function, converts a text containing ANSI codes
        into html statements.

        Args:
            text (str): Text to process.
            strip_ansi (bool, optional):

        Returns:
            text (str): Parsed text.

        """
        # parse everything to ansi first
        text = parse_ansi(text, strip_ansi=strip_ansi, xterm256=True, mxp=True)
        # convert all ansi to html
        result = re.sub(self.re_string, self.sub_text, text)
        result = re.sub(self.re_mxplink, self.sub_mxp_links, result)
        result = re.sub(self.re_mxpurl, self.sub_mxp_urls, result)
        result = self.remove_bells(result)
        result = self.format_styles(result)
        result = self.convert_linebreaks(result)
        result = self.remove_backspaces(result)
        result = self.convert_urls(result)
        # clean out eventual ansi that was missed
        ## result = parse_ansi(result, strip_ansi=True)

        return result


HTML_PARSER = TextToHTMLparser()


#
# Access function
#


def parse_html(string, strip_ansi=False, parser=HTML_PARSER):
    """
    Parses a string, replace ANSI markup with html
    """
    return parser.parse(string, strip_ansi=strip_ansi)



"""
The grid

This represents the full XYZ grid, which consists of

- 2D `Map`-objects parsed from Map strings and Map-legend components. Each represents one
  Z-coordinate or location.
- `Prototypes` for how to build each XYZ component into 'real' rooms and exits.
- Actual in-game rooms and exits, mapped to the game based on Map data.

The grid has three main functions:
- Building new rooms/exits from scratch based on one or more Maps.
- Updating the rooms/exits tied to an existing Map when the Map string
  of that map changes.
- Fascilitate communication between the in-game entities and their Map.


"""
from evennia.scripts.scripts import DefaultScript
from evennia.utils import logger
from evennia.utils.utils import variable_from_module

from .xymap import XYMap
from .xyzroom import XYZExit, XYZRoom


class XYZGrid(DefaultScript):
    """
    Main grid class. This organizes the Maps based on their name/Z-coordinate.

    """

    def at_script_creation(self):
        """
        What we store persistently is data used to create each map (the legends, names etc)

        """
        self.db.map_data = {}
        self.desc = "Manages maps for XYZ-grid"

    @property
    def grid(self):
        if self.ndb.grid is None:
            self.reload()
        return self.ndb.grid

    def get_map(self, zcoord):
        """
        Get a specific xymap.

        Args:
            zcoord (str): The name/zcoord of the xymap.

        Returns:
            XYMap: Or None if no map was found.

        """
        return self.grid.get(zcoord)

    def all_maps(self):
        """
        Get all xymaps stored in the grid.

        Returns:
            list: All initialized xymaps stored with this grid.

        """
        return list(self.grid.values())

    def log(self, msg):
        logger.log_info(f"|grid| {msg}")

    def get_room(self, xyz, **kwargs):
        """
        Get one or more room objects from XYZ coordinate.

        Args:
            xyz (tuple): X,Y,Z coordinate of room to fetch. '*' acts
            as wild cards.

        Returns:
            Queryset: A queryset of XYZRoom(s) found.

        Raises:
            XYZRoom.DoesNotExist: If room is not found.

        Notes:
            This assumes the room was previously built.

        """
        return XYZRoom.objects.filter_xyz(xyz=xyz, **kwargs)

    def get_exit(self, xyz, name="north", **kwargs):
        """
        Get one or more exit object at coordinate.

        Args:
            xyz (tuple): X,Y,Z coordinate of the room the
                exit leads out of. '*' acts as a wildcard.
            name (str): The full name of the exit, e.g. 'north' or 'northwest'.
                The '*' acts as a wild card.

        Returns:
            Queryset: A queryset of XYZExit(s) found.

        """
        kwargs["db_key"] = name
        return XYZExit.objects.filter_xyz_exit(xyz=xyz, **kwargs)

    def maps_from_module(self, module_path):
        """
        Load map data from module. The loader will look for a dict XYMAP_DATA or a list of
        XYMAP_DATA_LIST (a list of XYMAP_DATA dicts). Each XYMAP_DATA dict should contain
        `{"xymap": mapstring, "zcoord": mapname/zcoord, "legend": dict, "prototypes": dict}`.

        Args:
            module_path (module_path): A python-path to a module containing
                map data as either `XYMAP_DATA` or `XYMAP_DATA_LIST` variables.

        Returns:
            list: List of zero, one or more xy-map data dicts loaded from the module.

        """
        map_data_list = variable_from_module(module_path, "XYMAP_DATA_LIST")
        if not map_data_list:
            map_data_list = [variable_from_module(module_path, "XYMAP_DATA")]
        # inject the python path in the map data
        for mapdata in map_data_list:
            if not mapdata:
                self.log(f"Could not find or load map from {module_path}.")
                return
            mapdata["module_path"] = module_path
        return map_data_list

    def reload(self):
        """
        Reload and rebuild the grid. This is done on a server reload.

        """
        self.log("(Re)loading grid ...")
        self.ndb.grid = {}
        nmaps = 0
        loaded_mapdata = {}
        changed = []
        mapdata = self.db.map_data

        if not mapdata:
            self.db.mapdata = mapdata = {}

        # generate all Maps - this will also initialize their components
        # and bake any pathfinding paths (or load from disk-cache)
        for zcoord, old_mapdata in mapdata.items():
            self.log(f"Loading map '{zcoord}'...")

            # we reload the map from module
            new_mapdata = loaded_mapdata.get(zcoord)
            if not new_mapdata:
                if "module_path" in old_mapdata:
                    for mapdata in self.maps_from_module(old_mapdata["module_path"]):
                        loaded_mapdata[mapdata["zcoord"]] = mapdata
                else:
                    # nowhere to reload from - use what we have
                    loaded_mapdata[zcoord] = old_mapdata

                new_mapdata = loaded_mapdata.get(zcoord)

            if new_mapdata != old_mapdata:
                self.log(f" XYMap data for Z='{zcoord}' has changed.")
                changed.append(zcoord)

            xymap = XYMap(dict(new_mapdata), Z=zcoord, xyzgrid=self)
            xymap.parse()
            xymap.calculate_path_matrix()
            self.ndb.grid[zcoord] = xymap
            nmaps += 1

        # re-store changed data
        for zcoord in changed:
            self.db.map_data[zcoord] = loaded_mapdata[zcoord]

        # store
        self.log(f"Loaded and linked {nmaps} map(s).")
        self.ndb.loaded = True

    def add_maps(self, *mapdatas):
        """
        Add map or maps to the grid.

        Args:
            *mapdatas (dict): Each argument is a dict structure
                `{"map": <mapstr>, "legend": <legenddict>, "name": <name>,
                "prototypes": <dict-of-dicts>, "module_path": <str>}`. The `prototypes are
                coordinate-specific overrides for nodes/links on the map, keyed with their
                (X,Y) coordinate within that map. The `module_path` is injected automatically
                by self.maps_from_module.

        Raises:
            RuntimeError: If mapdata is malformed.

        """
        for mapdata in mapdatas:
            zcoord = mapdata.get("zcoord")
            if not zcoord is not None:
                raise RuntimeError("XYZGrid.add_map data must contain 'zcoord'.")

            self.db.map_data[zcoord] = mapdata

    def remove_map(self, *zcoords, remove_objects=True):
        """
        Remove an XYmap from the grid.

        Args:
            *zoords (str): The zcoords/XYmaps to remove.
            remove_objects (bool, optional): If the synced database objects (rooms/exits) should
                be removed alongside this map.
        """
        # from evennia import set_trace;set_trace()
        for zcoord in zcoords:
            if zcoord in self.db.map_data:
                self.db.map_data.pop(zcoord)
            if remove_objects:
                # we can't batch-delete because we want to run the .delete
                # method that also wipes exits and moves content to save locations
                for xyzroom in XYZRoom.objects.filter_xyz(xyz=("*", "*", zcoord)):
                    xyzroom.delete()
        self.reload()

    def delete(self):
        """
        Clear the entire grid, including database entities, then the grid too.

        """
        mapdata = self.db.map_data
        if mapdata:
            self.remove_map(*(zcoord for zcoord in self.db.map_data), remove_objects=True)
        super().delete()

    def spawn(self, xyz=("*", "*", "*"), directions=None):
        """
        Create/recreate/update the in-game grid based on the stored Maps or for a specific Map
        or coordinate.

        Args:
            xyz (tuple, optional): An (X,Y,Z) coordinate, where Z is the name of the map. `'*'`
                acts as a wildcard.
            directions (list, optional): A list of cardinal directions ('n', 'ne' etc).
                Spawn exits only the given direction. If unset, all needed directions are spawned.

        Examples:
            - `xyz=('*', '*', '*')` (default) - spawn/update all maps.
            - `xyz=(1, 3, 'foo')` - sync a specific element of map 'foo' only.
            - `xyz=('*', '*', 'foo') - sync all elements of map 'foo'
            - `xyz=(1, 3, '*') - sync all (1,3) coordinates on all maps (rarely useful)
            - `xyz=(1, 3, 'foo')`, `direction='ne'` - sync only the north-eastern exit
                out of the specific node on map 'foo'.

        """
        x, y, z = xyz
        wildcard = "*"

        if z == wildcard:
            xymaps = self.grid
        elif self.ndb.grid and z in self.ndb.grid:
            xymaps = {z: self.grid[z]}
        else:
            raise RuntimeError(f"The 'z' coordinate/name '{z}' is not found on the grid.")

        # first build all nodes/rooms
        for zcoord, xymap in xymaps.items():
            self.log(f"spawning/updating nodes for Z='{zcoord}' ...")
            xymap.spawn_nodes(xy=(x, y))

        # next build all links between nodes (including between maps)
        for zcoord, xymap in xymaps.items():
            self.log(f"spawning/updating links for Z='{zcoord}' ...")
            xymap.spawn_links(xy=(x, y), directions=directions)


def get_xyzgrid(print_errors=True):
    """
    Helper for getting the grid. This will create the XYZGrid global script if it didn't
    previously exist.

    Args:
        print_errors (bool, optional): Print errors directly to console rather than to log.

    """
    xyzgrid = XYZGrid.objects.all()
    if not xyzgrid:
        # create a new one
        xyzgrid, err = XYZGrid.create("XYZGrid")
        if err:
            raise RuntimeError(err)
        xyzgrid.reload()
        return xyzgrid
    elif len(xyzgrid) > 1:
        print(
            "Warning: More than one XYZGrid instance were found. This is an error and "
            "only the first one will be used. Delete the other one(s) manually."
        )
    xyzgrid = xyzgrid[0]
    try:
        if not xyzgrid.ndb.loaded:
            xyzgrid.reload()
    except Exception as err:
        # raise   # debug
        if print_errors:
            print(err)
        else:
            xyzgrid.log(str(err))
    return xyzgrid



"""
Helper to handle on-demand requests, allowing a system to change state only when a player or system
actually needs the information. This is a very efficient way to handle gradual changes, requiring
not computer resources until the state is actually needed.

For example, consider a flowering system, where a seed sprouts, grows and blooms over a certain time.
One _could_ implement this with e.g. a Script or a ticker that gradually moves the flower along
its stages of growth. But what if that flower is in a remote location, and no one is around to see it?
You are then wasting computational resources on something that no one is looking at.

The truth is that most of the time, players are not looking at most of the things in the game. They
_only_ need to know about which state the flower is in when they are actually looking at it, or
when they are in the same room as it (so it can be incorporated in the room description). This is
where on-demand handling comes in.

This is the basic principle, using the flowering system as an example.

1. Someone plants a seed in a room (could also be automated). The seed is in a "seedling" state.
    We store the time it was planted (this is the important bit).
2. A player enters the room or looks at the plant. We check the time it was planted, and calculate
  how much time has passed since it was planted. If enough time has passed, we change the state to
  "sprouting" and probably change its description to reflect this.
3. If a player looks at the plant and not enough time has passed, it keeps the last updated state.
4. Eventually, it will be bloom time, and the plant will change to a "blooming" state when the
   player looks.
5. If no player ever comes around to look at the plant, it will never change state, and if they show
   up after a long time, it may not show as a "wilted" state or be outright deleted when observed,
   since too long time has passed and the plant has died.

With a system like this you could have growing plants all over your world and computing usage would
only scale by how many players you have exploring your world. The players will not know the difference
between this and a system that is always running, but your server will thank you.

There is only one situation where this system is not ideal, and that is when a player should be
informed of the state change _even if they perform no action_. That is, even if they are just idling
in the room, they should get a message like 'the plant suddenly blooms' (or, more commonly, for
messages like 'you are feeling hungry'). For this you still probably need to use one of Evennia's
built-in timers or tickers instead. But most of the time you should really consider using on-demand
handling instead.

## Usage

```python

from evennia import ON_DEMAND_HANDLER

# create a new on-demand task

flower = create_object(Flower, key="rose")

ON_DEMAND_HANDLER.add_task(
    flower, category="flowering",
    stages={0: "seedling", 120: "sprouting",
            300: "blooming", 600: "wilted", 700: "dead"})

# later, when we want to check the state of the plant (e.g. in a command),

state = ON_DEMAND_HANDLER.get_stage("flowering", last_checked=plant.planted_time)

```


"""

from evennia.server.models import ServerConfig
from evennia.utils import logger
from evennia.utils.utils import is_iter

_RUNTIME = None

ON_DEMAND_HANDLER = None


class OnDemandTask:
    """
    Stores information about an on-demand task.

    Default property:
    - `default_stage_function (callable)`: This is called if no stage function is given in the stages dict.
        This is meant for changing the task itself (such as restarting it). Actual game code should
        be handled elsewhere, by checking this task. See the `stagefunc_*` static methods for examples
        of how to manipulate the task when a stage is reached.

    """

    # useful stage-functions. Use with OnDemandTask.endfunc_stop etc

    @staticmethod
    def runtime():
        """
        Wraps the gametime.runtime() function.

        Need to import here to avoid circular imports during server reboot.
        It's a callable to allow easier unit testing.

        """
        global _RUNTIME
        if not _RUNTIME:
            from evennia.utils.gametime import runtime as _RUNTIME
        return _RUNTIME()

    @staticmethod
    def stagefunc_loop(task):
        """
        Attach this to the last stage to have the task start over from
        the beginning

        Example:
            stages = {0: "seedling", 120: "flowering", 300: "dead", ("_loop",
            OnDemandTask.stagefunc_loop)}

            Note that the "respawn" state will never actually be visible as a state to
            the user, instead once it reaches this state, it will *immediately* loop
            and the new looped state will be shown and returned to the user. So it
            can an idea to mark that end state with a `_` just to indicate this fact.

        """

        now = OnDemandTask.runtime()
        original_start_time = (
            task.start_time
        )  # this can be set on start or previous call of this func
        dts = list(task.stages.keys())
        total_dt = max(dts) - min(dts)

        # figure out how many times we've looped since last start-time was set
        task.iterations += (now - original_start_time) // total_dt
        # figure out how far we are into the current loop.
        current_loop_time = (now - original_start_time) % total_dt
        # We need to adjust the start_time to the start of the current loop
        task.start_time = now - current_loop_time

    @staticmethod
    def stagefunc_bounce(task):
        """
        This endfunc will have the task reverse direction and go through the stages in
        reverse order. This stage-function must be placed at both 'ends' of the stage sequence
        for the bounce to continue indefinitely.

        Example:
            stages = {0: ("cool", OnDemandTask.stagefunc_bounce),
                      50: "lukewarm",
                      150: "warm",
                      300: "hot",
                      300: ("HOT!", OnDemandTask.stagefunc_bounce)}

        """
        now = OnDemandTask.runtime()
        original_start_time = (
            task.start_time
        )  # this can be set on start or previous call of this func
        dts = list(task.stages.keys())
        max_dt = max(dts)
        total_dt = max_dt - min(dts)
        task.iterations += (now - original_start_time) // total_dt
        current_loop_time = (now - original_start_time) % total_dt
        task.start_time = now - current_loop_time

        if task.iterations > 0:
            # reverse the stages
            stages = task.stages
            task.stages = {abs(k - max_dt): v for k, v in sorted(stages.items())}

    # default fallback stage function. This is called if no stage function is given in the stages dict.
    default_stage_function = None

    def __init__(self, key, category, stages=None, autostart=True):
        """
        Args:
            key (str): A unique identifier for the task.
            stages (dict, optional): A dictionary `{dt: str}` or `{int or float: (str, callable)}`
                of time-deltas (in seconds) and the stage name they represent. If the value is a
                tuple, the first element is the name of the stage and the second is a callable
                that will be called when that stage is *first* reached. Warning: This callable
                is *only* triggered if the stage is actually checked/retrieved while the task is
                in that stage checks - it's _not_ guaranteed to be called, even if the task
                time-wise goes through all its stages. Each callable must be picklable (so normally
                it should be a stand-alone function), and takes one argument - this OnDemandTask,
                which it can be modified in-place as needed.  This can be used to loop a task or do
                other changes to the task.
            autostart (bool, optional): If `last_checked` is `None`, and this is `False`, then the
                time will not start counting until the first call of `get_dt` or `get_stage`. If
                `True`, creating the task will immediately make a hidden check and start the timer.

        Examples:

           stages = {0: "seedling",
                     120: "sprouting",
                     300: "blooming",
                     600: "wilted",
                     700: "dead"
                    }

        """
        self.key = key
        self.category = category
        self.start_time = None
        self.last_stage = None
        self.iterations = 0  # only used with looping staging functions

        self.stages = None

        if isinstance(stages, dict):
            # sort the stages by ending time, inserting each state as {dt: (statename, callable)}
            _stages = {}
            for dt, tup in stages.items():
                if is_iter(tup):
                    if len(tup) != 2:
                        raise ValueError(
                            "Each stage must be a tuple (name, callable) or a name-string."
                        )
                    if not callable(tup[1]):
                        raise ValueError(
                            "The second element of each stage-tuple must be a callable."
                        )
                else:
                    tup = (tup, None)
                _stages[dt] = tup

            self.stages = {dt: tup for dt, tup in sorted(_stages.items(), reverse=True)}

        self.check(autostart=autostart)

    def __str__(self):
        """Note that we don't check the state here"""
        # we visualize stages with ascending key order
        dt, stage = self.check(autostart=False)
        return f"OnDemandTask({self.key}[{self.category}] (dt={dt}s), stage={stage})"

    def __eq__(self, other):
        if not isinstance(other, OnDemandTask):
            return False
        return (self.key, self.category) == (other.key, other.category)

    def check(self, autostart=True):
        """
        Check the current stage of the task and return the time-delta to the next stage.

        Args:
            autostart (bool, optional): If this is set, and the task has not been started yet,
                it will be started by this check. This is mainly used internally.

        Returns:
            tuple: A tuple (dt, stage) where `dt` is the time-delta (in seconds) since the test
            started (or since it started its latest iteration). and `stage` is the name of the
            current stage. If no stages are defined, `stage` will always be `None`. Use `get_dt` and
            `get_stage` to get only one of these values.

        """

        def _find_stage(delta_dt, _rerun=False):
            if not self.stages:
                return None

            for dt, (stage, stage_func) in self.stages.items():
                if delta_dt < dt:
                    continue

                if autostart and stage != self.last_stage and not _rerun:
                    self.last_stage = stage

                    if stage_func:
                        try:
                            stage_func(self)
                        except Exception as err:
                            logger.log_trace(
                                f"Error getting stage of on-demand task {self} "
                                f"(last_stage: {self.last_stage}, trying to call stage-func "
                                f"{stage_func}: {err}"
                            )
                        else:
                            # rerun the check in case the endfunc changed things
                            return _find_stage(delta_dt, _rerun=True)
                return stage

        def _find_dt(self, autostart=False):
            if self.start_time is None:
                if autostart:
                    # start timer
                    self.start_time = now
                dt = 0
            else:
                dt = now - self.start_time
            return dt

        now = OnDemandTask.runtime()

        dt = _find_dt(self, autostart=autostart)

        # we must always fetch the stage since a stage_func may fire
        stage = _find_stage(dt)

        # need to fetch dt again in case stage_func changed it
        dt = _find_dt(self, autostart=autostart)

        return dt, stage

    def get_dt(self):
        """
        Get the time-delta since last check.

        Returns:
            int: The time since the last check, or 0 if this is the first time the task is checked.

        """

        return self.check()[0]

    def get_stage(self):
        """
        Get the current stage of the task. If no stage was given, this will return `None` but
        still update the last_checked time.

        Returns:
            str or None: The current stage of the task, or `None` if no stages are set.

        """
        return self.check()[1]


class OnDemandHandler:
    """
    A singleton handler for managing on-demand state changes. Its main function is to persistently
    track the time (in seconds) between a state change and the next. How you make use of this
    information is up to your particular system.

    Contrary to just using the `time` module, this will also account for server restarts.

    """

    def __init__(self):
        self.tasks = dict()

    def load(self):
        """
        Load the on-demand timers from ServerConfig storage.

        This should be automatically called when Evennia starts.

        """
        self.tasks = ServerConfig.objects.conf("on_demand_timers", default=dict)

    def save(self):
        """
        Save the on-demand timers to ServerConfig storage. Should be called when Evennia shuts down.

        """
        ServerConfig.objects.conf("on_demand_timers", self.tasks)

    def _build_key(self, key, category):
        """
        Build a unique key for the task.

        Args:
            key (str, callable, OnDemandTask or Object): The task key. If callable, it will be
                called without arguments. If an Object, will be converted to a string. If
                an `OnDemandTask`, then all other arguments are ignored and the task will be used
                to build the internal storage key.
            category (str or callable): The task category. If callable, it will be called without arguments.

        Returns:
            tuple (str, str or None): The unique key.

        """
        if isinstance(key, OnDemandTask):
            return (key.key, key.category)

        return (
            str(key() if callable(key) else key),
            category() if callable(category) else str(category) if category is not None else None,
        )

    def add(self, key, category=None, stages=None, autostart=True):
        """
        Add a new on-demand task.

        Args:
            key (str, callable, OnDemandTask or Object): A unique identifier for the task. If this
                is a callable, it will be called without arguments. If a db-Object, it will be
                converted to a string representation (which will include its (#dbref). If an
                `OnDemandTask`, then all other arguments are ignored and the task is simply added
                as-is.
            category (str or callable, optional): A category to group the task under. If given, it
                must also be given when checking the task.
            stages (dict, optional): A dictionary {dt: str}, of time-deltas (in seconds) and the
                stage which should be entered after that much time has passed.  autostart (bool,
            optional): If `True`, creating the task will immediately make a hidden
                check and start the timer.

        Returns:
            OnDemandTask: The created task (or the same that was added, if given an `OnDemandTask`
                as a `key`).  Use `task.get_dt()` and `task.get_stage()` to get data from it manually.

        """
        if isinstance(key, OnDemandTask):
            self.tasks[self._build_key(key.key, key.category)] = key
            return key
        task = OnDemandTask(key, category, stages, autostart=autostart)
        self.tasks[self._build_key(key, category)] = task
        return task

    def batch_add(self, *tasks):
        """
        Add multiple on-demand tasks at once.

        Args:
            *tasks (OnDemandTask): A set of OnDemandTasks to add.

        """
        for task in tasks:
            self.tasks[self._build_key(task.key, task.category)] = task

    def remove(self, key, category=None):
        """
        Remove an on-demand task.

        Args:
            key (str, callable, OnDemandTask or Object): The unique identifier for the task. If a callable, will
                be called without arguments. If an Object, will be converted to a string. If an `OnDemandTask`,
                then all other arguments are ignored and the task will be used to identify the task to remove.
            category (str or callable, optional): The category of the task.

        Returns:
            OnDemandTask or None: The removed task, or `None` if no task was found.

        """
        return self.tasks.pop(self._build_key(key, category), None)

    def batch_remove(self, *keys, category=None):
        """
        Remove multiple on-demand tasks at once, potentially within a given category.

        Args:
            *keys (str, callable, OnDemandTask or Object): The unique identifiers for the tasks. If
                a callable, will be called without arguments. If an Object, will be converted to a
                string. If an `OnDemandTask`, then all other arguments are ignored and the task will
                be used to identify the task to remove.
            category (str or callable, optional): The category of the tasks.

        """
        for key in keys:
            self.remove(key, category=category)

    def all(self, category=None, all_on_none=True):
        """
        Get all on-demand tasks.

        Args:
            category (str, optional): The category of the tasks.
            all_on_none (bool, optional): Determines what to return if `category` is `None`.
                If `True`, all tasks will be returned. If `False`, only tasks without a category
                will be returned.

        Returns:
            dict: A dictionary of all on-demand task, on the form `{(key, category): task), ...}`.
            Use `task.get_dt()` or `task.get_stage()` to get the time-delta or stage of each task
            manually.

        """
        if category is None and all_on_none:
            # return all
            return self.tasks

        # filter by category (treat no-category as its own category)
        return {keytuple: task for keytuple, task in self.tasks.items() if keytuple[1] == category}

    def clear(self, category=None, all_on_none=True):
        """
        Clear all on-demand tasks.

        Args:
            category (str, optional): The category of the tasks to clear. What `None` means is determined
                by the `all_on_none` kwarg.
            all_on_none (bool, optional): Determines what to clear if `category` is `None`. If `True`,
                clear all tasks, if `False`, only clear tasks with no category.

        """
        if category is None and all_on_none:
            # clear all
            self.tasks = {}

        # filter and clear only those matching the category
        self.tasks = {
            keytuple: task for keytuple, task in self.tasks.items() if keytuple[1] != category
        }

    def get(self, key, category=None):
        """
        Get an on-demand task. This will _not_ check it.

        Args:
            key (str, callable, OnDemandTask or Object): The unique identifier for the task. If a
                callable, will be called without arguments. If an Object, will be converted to a string.
                If an `OnDemandTask`, then all other arguments are ignored and the task will be used
                (only useful to check the task is the same).

            category (str, optional): The category of the task. If unset, this will only return
                tasks with no category.

        Returns:
            OnDemandTask or None: The task, or `None` if no task was found.

        """
        return self.tasks.get(self._build_key(key, category))

    def get_dt(self, key, category=None):
        """
        Get the time-delta since the task started.

        Args:
            key (str, callable, OnDemandTask or Object): The unique identifier for the task. If a
                callable, will be called without arguments. If an Object, will be converted to a string.
                If an `OnDemandTask`, then all other arguments are ignored and the task will be used
                to identify the task to get the time-delta from.

        Returns:
            int or None: The time since the last check, or `None` if no task was found.

        """
        task = self.get(key, category)
        return task.get_dt() if task else None

    def get_stage(self, key, category=None):
        """
        Get the current stage of an on-demand task.

        Args:
            key (str, callable, OnDemandTask or Object): The unique identifier for the task. If a
                callable, will be called without arguments. If an Object, will be converted to a string.
                If an `OnDemandTask`, then all other arguments are ignored and the task will be used
                to identify the task to get the stage from.

        Returns:
            str or None: The current stage of the task, or `None` if no task was found.

        """
        task = self.get(key, category)
        return task.get_stage() if task else None


# Create singleton
ON_DEMAND_HANDLER = OnDemandHandler()



"""
Account

The account class is an extension of the default Django user class,
and is customized for the needs of Evennia.

We use the Account to store a more mud-friendly style of permission
system as well as to allow the admin more flexibility by storing
attributes on the Account.  Within the game we should normally use the
Account manager's methods to create users so that permissions are set
correctly.

To make the Account model more flexible for your own game, it can also
persistently store attributes of its own. This is ideal for extra
account info and OOC account configuration variables etc.

"""
from django.conf import settings
from django.contrib.auth.models import AbstractUser
from django.db import models
from django.utils.encoding import smart_str

from evennia.accounts.manager import AccountDBManager
from evennia.server.signals import SIGNAL_ACCOUNT_POST_RENAME
from evennia.typeclasses.models import TypedObject
from evennia.utils.utils import make_iter

__all__ = ("AccountDB",)

_GA = object.__getattribute__
_SA = object.__setattr__
_DA = object.__delattr__

_TYPECLASS = None


# ------------------------------------------------------------
#
# AccountDB
#
# ------------------------------------------------------------


class AccountDB(TypedObject, AbstractUser):
    """
    This is a special model using Django's 'profile' functionality
    and extends the default Django User model. It is defined as such
    by use of the variable AUTH_PROFILE_MODULE in the settings.
    One accesses the fields/methods. We try use this model as much
    as possible rather than User, since we can customize this to
    our liking.

    The TypedObject supplies the following (inherited) properties:

      - key - main name
      - typeclass_path - the path to the decorating typeclass
      - typeclass - auto-linked typeclass
      - date_created - time stamp of object creation
      - permissions - perm strings
      - dbref - #id of object
      - db - persistent attribute storage
      - ndb - non-persistent attribute storage

    The AccountDB adds the following properties:

      - is_connected - If any Session is currently connected to this Account
      - name - alias for user.username
      - sessions - sessions connected to this account
      - is_superuser - bool if this account is a superuser
      - is_bot - bool if this account is a bot and not a real account

    """

    #
    # AccountDB Database model setup
    #
    # inherited fields (from TypedObject):
    # db_key, db_typeclass_path, db_date_created, db_permissions

    # store a connected flag here too, not just in sessionhandler.
    # This makes it easier to track from various out-of-process locations
    db_is_connected = models.BooleanField(
        default=False,
        verbose_name="is_connected",
        help_text="If player is connected to game or not",
    )
    # database storage of persistant cmdsets.
    db_cmdset_storage = models.CharField(
        "cmdset",
        max_length=255,
        null=True,
        help_text=(
            "optional python path to a cmdset class. If creating a Character, this will "
            "default to settings.CMDSET_CHARACTER."
        ),
    )
    # marks if this is a "virtual" bot account object
    db_is_bot = models.BooleanField(
        default=False, verbose_name="is_bot", help_text="Used to identify irc/rss bots"
    )

    # Database manager
    objects = AccountDBManager()

    # defaults
    __defaultclasspath__ = "evennia.accounts.accounts.DefaultAccount"
    __applabel__ = "accounts"
    __settingsclasspath__ = settings.BASE_SCRIPT_TYPECLASS

    class Meta:
        verbose_name = "Account"

    # cmdset_storage property
    # This seems very sensitive to caching, so leaving it be for now /Griatch
    # @property
    def __cmdset_storage_get(self):
        """
        Getter. Allows for value = self.name. Returns a list of cmdset_storage.
        """
        storage = self.db_cmdset_storage
        # we need to check so storage is not None
        return [path.strip() for path in storage.split(",")] if storage else []

    # @cmdset_storage.setter
    def __cmdset_storage_set(self, value):
        """
        Setter. Allows for self.name = value. Stores as a comma-separated
        string.
        """
        _SA(self, "db_cmdset_storage", ",".join(str(val).strip() for val in make_iter(value)))
        _GA(self, "save")()

    # @cmdset_storage.deleter
    def __cmdset_storage_del(self):
        "Deleter. Allows for del self.name"
        _SA(self, "db_cmdset_storage", None)
        _GA(self, "save")()

    cmdset_storage = property(__cmdset_storage_get, __cmdset_storage_set, __cmdset_storage_del)

    #
    # property/field access
    #

    def __str__(self):
        return smart_str(f"{self.name}(account {self.dbid})")

    def __repr__(self):
        return f"{self.name}(account#{self.dbid})"

    # @property
    def __username_get(self):
        return self.username

    def __username_set(self, value):
        old_name = self.username
        self.username = value
        self.save(update_fields=["username"])
        SIGNAL_ACCOUNT_POST_RENAME.send(self, old_name=old_name, new_name=value)

    def __username_del(self):
        del self.username

    # aliases
    name = property(__username_get, __username_set, __username_del)
    key = property(__username_get, __username_set, __username_del)

    # @property
    def __uid_get(self):
        "Getter. Retrieves the user id"
        return self.id

    def __uid_set(self, value):
        raise Exception("User id cannot be set!")

    def __uid_del(self):
        raise Exception("User id cannot be deleted!")

    uid = property(__uid_get, __uid_set, __uid_del)



# -*- coding: utf-8 -*-


from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("objects", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("typeclasses", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="ScriptDB",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID", serialize=False, auto_created=True, primary_key=True
                    ),
                ),
                ("db_key", models.CharField(max_length=255, verbose_name="key", db_index=True)),
                (
                    "db_typeclass_path",
                    models.CharField(
                        help_text="this defines what 'type' of entity this is. This variable holds a Python path to a module with a valid Evennia Typeclass.",
                        max_length=255,
                        null=True,
                        verbose_name="typeclass",
                    ),
                ),
                (
                    "db_date_created",
                    models.DateTimeField(auto_now_add=True, verbose_name="creation date"),
                ),
                (
                    "db_lock_storage",
                    models.TextField(
                        help_text="locks limit access to an entity. A lock is defined as a 'lock string' on the form 'type:lockfunctions', defining what functionality is locked and how to determine access. Not defining a lock means no access is granted.",
                        verbose_name="locks",
                        blank=True,
                    ),
                ),
                ("db_desc", models.CharField(max_length=255, verbose_name="desc", blank=True)),
                (
                    "db_interval",
                    models.IntegerField(
                        default=-1,
                        help_text="how often to repeat script, in seconds. -1 means off.",
                        verbose_name="interval",
                    ),
                ),
                (
                    "db_start_delay",
                    models.BooleanField(
                        default=False,
                        help_text="pause interval seconds before starting.",
                        verbose_name="start delay",
                    ),
                ),
                (
                    "db_repeats",
                    models.IntegerField(
                        default=0, help_text="0 means off.", verbose_name="number of repeats"
                    ),
                ),
                (
                    "db_persistent",
                    models.BooleanField(default=False, verbose_name="survive server reboot"),
                ),
                ("db_is_active", models.BooleanField(default=False, verbose_name="script active")),
                (
                    "db_attributes",
                    models.ManyToManyField(
                        help_text="attributes on this object. An attribute can hold any pickle-able python object (see docs for special cases).",
                        to="typeclasses.Attribute",
                        null=True,
                    ),
                ),
                (
                    "db_obj",
                    models.ForeignKey(
                        blank=True,
                        to="objects.ObjectDB",
                        on_delete=models.CASCADE,
                        help_text="the object to store this script on, if not a global script.",
                        null=True,
                        verbose_name="scripted object",
                    ),
                ),
                (
                    "db_account",
                    models.ForeignKey(
                        blank=True,
                        to=settings.AUTH_USER_MODEL,
                        on_delete=models.CASCADE,
                        help_text="the account to store this script on (should not be set if obj is set)",
                        null=True,
                        verbose_name="scripted account",
                    ),
                ),
                (
                    "db_tags",
                    models.ManyToManyField(
                        help_text="tags on this object. Tags are simple string markers to identify, group and alias objects.",
                        to="typeclasses.Tag",
                        null=True,
                    ),
                ),
            ],
            options={"verbose_name": "Script"},
            bases=(models.Model,),
        )
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-06-25 18:12


from django.db import migrations

# Migration of old nick format to new


def update_nicks(apps, schema_editor):
    Attribute = apps.get_model("typeclasses", "Attribute")
    for nick in Attribute.objects.filter(db_attrtype="nick"):
        try:
            _, _, _, _ = nick.db_value
        except (TypeError, ValueError):
            # old setup, we store it in the new format - old uses its own key
            # as regex to keep the old matching (and has no $-type args)
            nick.db_value = (nick.db_key, nick.db_strvalue, nick.db_key, nick.db_strvalue)
            nick.save()


class Migration(migrations.Migration):

    dependencies = [
        ("typeclasses", "0004_auto_20151101_1759"),
        ("comms", "0010_auto_20161206_1912"),
        ("help", "0001_initial"),
    ]

    operations = [migrations.RunPython(update_nicks)]



"""
Simple turn-based combat system with equipment

Contrib - Tim Ashley Jenkins 2017, Refactor by Griatch 2022

This is a version of the 'turnbattle' contrib with a basic system for
weapons and armor implemented. Weapons can have unique damage ranges
and accuracy modifiers, while armor can reduce incoming damage and
change one's chance of getting hit. The 'wield' command is used to
equip weapons and the 'don' command is used to equip armor.

Some prototypes are included at the end of this module - feel free to
copy them into your game's prototypes.py module in your 'world' folder
and create them with the @spawn command. (See the tutorial for using
the @spawn command for details.)

For the example equipment given, heavier weapons deal more damage
but are less accurate, while light weapons are more accurate but
deal less damage. Similarly, heavy armor reduces incoming damage by
a lot but increases your chance of getting hit, while light armor is
easier to dodge in but reduces incoming damage less. Light weapons are
more effective against lightly armored opponents and heavy weapons are
more damaging against heavily armored foes, but heavy weapons and armor
are slightly better than light weapons and armor overall.

This is a fairly bare implementation of equipment that is meant to be
expanded to fit your game - weapon and armor slots, damage types and
damage bonuses, etc. should be fairly simple to implement according to
the rules of your preferred system or the needs of your own game.

To install and test, import this module's TBEquipCharacter object into
your game's character.py module:

    from evennia.contrib.game_systems.turnbattle.tb_equip import TBEquipCharacter

And change your game's character typeclass to inherit from TBEquipCharacter
instead of the default:

    class Character(TBEquipCharacter):

Next, import this module into your default_cmdsets.py module:

    from evennia.contrib.game_systems.turnbattle import tb_equip

And add the battle command set to your default command set:

    #
    # any commands you add below will overload the default ones.
    #
    self.add(tb_equip.BattleCmdSet())

This module is meant to be heavily expanded on, so you may want to copy it
to your game's 'world' folder and modify it there rather than importing it
in your game and using it as-is.
"""

from random import randint

from evennia import Command, DefaultObject, default_cmds

from . import tb_basic

"""
----------------------------------------------------------------------------
OPTIONS
----------------------------------------------------------------------------
"""

TURN_TIMEOUT = 30  # Time before turns automatically end, in seconds
ACTIONS_PER_TURN = 1  # Number of actions allowed per turn

"""
----------------------------------------------------------------------------
COMBAT FUNCTIONS START HERE
----------------------------------------------------------------------------
"""


class EquipmentCombatRules(tb_basic.BasicCombatRules):
    """
    Has all the methods of the basic combat, with the addition of equipment.

    """

    def get_attack(self, attacker, defender):
        """
        Returns a value for an attack roll.

        Args:
            attacker (obj): Character doing the attacking
            defender (obj): Character being attacked

        Returns:
            attack_value (int): Attack roll value, compared against a defense value
                to determine whether an attack hits or misses.

        Notes:
            In this example, a weapon's accuracy bonus is factored into the attack
            roll. Lighter weapons are more accurate but less damaging, and heavier
            weapons are less accurate but deal more damage. Of course, you can
            change this paradigm completely in your own game.
        """
        # Start with a roll from 1 to 100.
        attack_value = randint(1, 100)
        accuracy_bonus = 0
        # If armed, add weapon's accuracy bonus.
        if attacker.db.wielded_weapon:
            weapon = attacker.db.wielded_weapon
            accuracy_bonus += weapon.db.accuracy_bonus
        # If unarmed, use character's unarmed accuracy bonus.
        else:
            accuracy_bonus += attacker.db.unarmed_accuracy
        # Add the accuracy bonus to the attack roll.
        attack_value += accuracy_bonus
        return attack_value

    def get_defense(self, attacker, defender):
        """
        Returns a value for defense, which an attack roll must equal or exceed in order
        for an attack to hit.

        Args:
            attacker (obj): Character doing the attacking
            defender (obj): Character being attacked

        Returns:
            defense_value (int): Defense value, compared against an attack roll
                to determine whether an attack hits or misses.

        Notes:
            Characters are given a default defense value of 50 which can be
            modified up or down by armor. In this example, wearing armor actually
            makes you a little easier to hit, but reduces incoming damage.
        """
        # Start with a defense value of 50 for a 50/50 chance to hit.
        defense_value = 50
        # Modify this value based on defender's armor.
        if defender.db.worn_armor:
            armor = defender.db.worn_armor
            defense_value += armor.db.defense_modifier
        return defense_value

    def get_damage(self, attacker, defender):
        """
        Returns a value for damage to be deducted from the defender's HP after abilities
        successful hit.

        Args:
            attacker (obj): Character doing the attacking
            defender (obj): Character being damaged

        Returns:
            damage_value (int): Damage value, which is to be deducted from the defending
                character's HP.

        Notes:
            Damage is determined by the attacker's wielded weapon, or the attacker's
            unarmed damage range if no weapon is wielded. Incoming damage is reduced
            by the defender's armor.
        """
        damage_value = 0
        # Generate a damage value from wielded weapon if armed
        if attacker.db.wielded_weapon:
            weapon = attacker.db.wielded_weapon
            # Roll between minimum and maximum damage
            damage_value = randint(weapon.db.damage_range[0], weapon.db.damage_range[1])
        # Use attacker's unarmed damage otherwise
        else:
            damage_value = randint(
                attacker.db.unarmed_damage_range[0], attacker.db.unarmed_damage_range[1]
            )
        # If defender is armored, reduce incoming damage
        if defender.db.worn_armor:
            armor = defender.db.worn_armor
            damage_value -= armor.db.damage_reduction
        # Make sure minimum damage is 0
        if damage_value < 0:
            damage_value = 0
        return damage_value

    def resolve_attack(self, attacker, defender, attack_value=None, defense_value=None):
        """
        Resolves an attack and outputs the result.

        Args:
            attacker (obj): Character doing the attacking
            defender (obj): Character being attacked

        Notes:
            Even though the attack and defense values are calculated
            extremely simply, they are separated out into their own functions
            so that they are easier to expand upon.
        """
        # Get the attacker's weapon type to reference in combat messages.
        attackers_weapon = "attack"
        if attacker.db.wielded_weapon:
            weapon = attacker.db.wielded_weapon
            attackers_weapon = weapon.db.weapon_type_name
        # Get an attack roll from the attacker.
        if not attack_value:
            attack_value = self.get_attack(attacker, defender)
        # Get a defense value from the defender.
        if not defense_value:
            defense_value = self.get_defense(attacker, defender)
        # If the attack value is lower than the defense value, miss. Otherwise, hit.
        if attack_value < defense_value:
            attacker.location.msg_contents(
                "%s's %s misses %s!" % (attacker, attackers_weapon, defender)
            )
        else:
            damage_value = self.get_damage(attacker, defender)  # Calculate damage value.
            # Announce damage dealt and apply damage.
            if damage_value > 0:
                attacker.location.msg_contents(
                    "%s's %s strikes %s for %i damage!"
                    % (attacker, attackers_weapon, defender, damage_value)
                )
            else:
                attacker.location.msg_contents(
                    "%s's %s bounces harmlessly off %s!" % (attacker, attackers_weapon, defender)
                )
            self.apply_damage(defender, damage_value)
            # If defender HP is reduced to 0 or less, call at_defeat.
            if defender.db.hp <= 0:
                self.at_defeat(defender)


COMBAT_RULES = EquipmentCombatRules()

"""
----------------------------------------------------------------------------
SCRIPTS START HERE
----------------------------------------------------------------------------
"""


class TBEquipTurnHandler(tb_basic.TBBasicTurnHandler):
    """
    This is the script that handles the progression of combat through turns.
    On creation (when a fight is started) it adds all combat-ready characters
    to its roster and then sorts them into a turn order. There can only be one
    fight going on in a single room at a time, so the script is assigned to a
    room as its object.

    Fights persist until only one participant is left with any HP or all
    remaining participants choose to end the combat with the 'disengage' command.
    """

    rules = COMBAT_RULES


"""
----------------------------------------------------------------------------
TYPECLASSES START HERE
----------------------------------------------------------------------------
"""


class TBEWeapon(DefaultObject):
    """
    A weapon which can be wielded in combat with the 'wield' command.

    """

    rules = COMBAT_RULES

    def at_object_creation(self):
        """
        Called once, when this object is first created. This is the
        normal hook to overload for most object types.
        """
        self.db.damage_range = (15, 25)  # Minimum and maximum damage on hit
        self.db.accuracy_bonus = 0  # Bonus to attack rolls (or penalty if negative)
        self.db.weapon_type_name = (
            "weapon"  # Single word for weapon - I.E. "dagger", "staff", "scimitar"
        )

    def at_drop(self, dropper):
        """
        Stop being wielded if dropped.
        """
        if dropper.db.wielded_weapon == self:
            dropper.db.wielded_weapon = None
            dropper.location.msg_contents("%s stops wielding %s." % (dropper, self))

    def at_give(self, giver, getter):
        """
        Stop being wielded if given.
        """
        if giver.db.wielded_weapon == self:
            giver.db.wielded_weapon = None
            giver.location.msg_contents("%s stops wielding %s." % (giver, self))


class TBEArmor(DefaultObject):
    """
    A set of armor which can be worn with the 'don' command.
    """

    def at_object_creation(self):
        """
        Called once, when this object is first created. This is the
        normal hook to overload for most object types.
        """
        self.db.damage_reduction = 4  # Amount of incoming damage reduced by armor
        self.db.defense_modifier = (
            -4
        )  # Amount to modify defense value (pos = harder to hit, neg = easier)

    def at_pre_drop(self, dropper):
        """
        Can't drop in combat.
        """
        if self.rules.is_in_combat(dropper):
            dropper.msg("You can't doff armor in a fight!")
            return False
        return True

    def at_drop(self, dropper):
        """
        Stop being wielded if dropped.
        """
        if dropper.db.worn_armor == self:
            dropper.db.worn_armor = None
            dropper.location.msg_contents("%s removes %s." % (dropper, self))

    def at_pre_give(self, giver, getter):
        """
        Can't give away in combat.
        """
        if self.rules.is_in_combat(giver):
            dropper.msg("You can't doff armor in a fight!")
            return False
        return True

    def at_give(self, giver, getter):
        """
        Stop being wielded if given.
        """
        if giver.db.worn_armor == self:
            giver.db.worn_armor = None
            giver.location.msg_contents("%s removes %s." % (giver, self))


class TBEquipCharacter(tb_basic.TBBasicCharacter):
    """
    A character able to participate in turn-based combat. Has attributes for current
    and maximum HP, and access to combat commands.
    """

    def at_object_creation(self):
        """
        Called once, when this object is first created. This is the
        normal hook to overload for most object types.
        """
        self.db.max_hp = 100  # Set maximum HP to 100
        self.db.hp = self.db.max_hp  # Set current HP to maximum
        self.db.wielded_weapon = None  # Currently used weapon
        self.db.worn_armor = None  # Currently worn armor
        self.db.unarmed_damage_range = (5, 15)  # Minimum and maximum unarmed damage
        self.db.unarmed_accuracy = 30  # Accuracy bonus for unarmed attacks

        """
        Adds attributes for a character's current and maximum HP.
        We're just going to set this value at '100' by default.

        You may want to expand this to include various 'stats' that
        can be changed at creation and factor into combat calculations.
        """


"""
----------------------------------------------------------------------------
COMMANDS START HERE
----------------------------------------------------------------------------
"""


class CmdFight(tb_basic.CmdFight):
    """
    Starts a fight with everyone in the same room as you.

    Usage:
      fight

    When you start a fight, everyone in the room who is able to
    fight is added to combat, and a turn order is randomly rolled.
    When it's your turn, you can attack other characters.
    """

    key = "fight"
    help_category = "combat"

    rules = COMBAT_RULES
    command_handler_class = TBEquipTurnHandler


class CmdAttack(tb_basic.CmdAttack):
    """
    Attacks another character.

    Usage:
      attack <target>

    When in a fight, you may attack another character. The attack has
    a chance to hit, and if successful, will deal damage.
    """

    key = "attack"
    help_category = "combat"

    rules = COMBAT_RULES


class CmdPass(tb_basic.CmdPass):
    """
    Passes on your turn.

    Usage:
      pass

    When in a fight, you can use this command to end your turn early, even
    if there are still any actions you can take.
    """

    key = "pass"
    aliases = ["wait", "hold"]
    help_category = "combat"

    rules = COMBAT_RULES


class CmdDisengage(tb_basic.CmdDisengage):
    """
    Passes your turn and attempts to end combat.

    Usage:
      disengage

    Ends your turn early and signals that you're trying to end
    the fight. If all participants in a fight disengage, the
    fight ends.
    """

    key = "disengage"
    aliases = ["spare"]
    help_category = "combat"

    rules = COMBAT_RULES


class CmdRest(tb_basic.CmdRest):
    """
    Recovers damage.

    Usage:
      rest

    Resting recovers your HP to its maximum, but you can only
    rest if you're not in a fight.
    """

    key = "rest"
    help_category = "combat"

    rules = COMBAT_RULES


class CmdCombatHelp(tb_basic.CmdCombatHelp):
    """
    View help or a list of topics

    Usage:
      help <topic or command>
      help list
      help all

    This will search for help on commands and other
    topics related to the game.
    """

    rules = COMBAT_RULES


class CmdWield(Command):
    """
    Wield a weapon you are carrying

    Usage:
      wield <weapon>

    Select a weapon you are carrying to wield in combat. If
    you are already wielding another weapon, you will switch
    to the weapon you specify instead. Using this command in
    combat will spend your action for your turn. Use the
    "unwield" command to stop wielding any weapon you are
    currently wielding.
    """

    key = "wield"
    help_category = "combat"

    rules = COMBAT_RULES

    def func(self):
        """
        This performs the actual command.
        """
        # If in combat, check to see if it's your turn.
        if self.rules.is_in_combat(self.caller):
            if not self.rules.is_turn(self.caller):
                self.caller.msg("You can only do that on your turn.")
                return
        if not self.args:
            self.caller.msg("Usage: wield <obj>")
            return
        weapon = self.caller.search(self.args, candidates=self.caller.contents)
        if not weapon:
            return
        if not weapon.is_typeclass(
            "evennia.contrib.game_systems.turnbattle.tb_equip.TBEWeapon", exact=True
        ):
            self.caller.msg("That's not a weapon!")
            # Remember to update the path to the weapon typeclass if you move this module!
            return

        if not self.caller.db.wielded_weapon:
            self.caller.db.wielded_weapon = weapon
            self.caller.location.msg_contents("%s wields %s." % (self.caller, weapon))
        else:
            old_weapon = self.caller.db.wielded_weapon
            self.caller.db.wielded_weapon = weapon
            self.caller.location.msg_contents(
                "%s lowers %s and wields %s." % (self.caller, old_weapon, weapon)
            )
        # Spend an action if in combat.
        if self.rules.is_in_combat(self.caller):
            self.rules.spend_action(self.caller, 1, action_name="wield")  # Use up one action.


class CmdUnwield(Command):
    """
    Stop wielding a weapon.

    Usage:
      unwield

    After using this command, you will stop wielding any
    weapon you are currently wielding and become unarmed.
    """

    key = "unwield"
    help_category = "combat"

    rules = COMBAT_RULES

    def func(self):
        """
        This performs the actual command.
        """
        # If in combat, check to see if it's your turn.
        if self.rules.is_in_combat(self.caller):
            if not self.rules.is_turn(self.caller):
                self.caller.msg("You can only do that on your turn.")
                return
        if not self.caller.db.wielded_weapon:
            self.caller.msg("You aren't wielding a weapon!")
        else:
            old_weapon = self.caller.db.wielded_weapon
            self.caller.db.wielded_weapon = None
            self.caller.location.msg_contents("%s lowers %s." % (self.caller, old_weapon))


class CmdDon(Command):
    """
    Don armor that you are carrying

    Usage:
      don <armor>

    Select armor to wear in combat. You can't use this
    command in the middle of a fight. Use the "doff"
    command to remove any armor you are wearing.
    """

    key = "don"
    help_category = "combat"

    rules = COMBAT_RULES

    def func(self):
        """
        This performs the actual command.
        """
        # Can't do this in combat
        if self.rules.is_in_combat(self.caller):
            self.caller.msg("You can't don armor in a fight!")
            return
        if not self.args:
            self.caller.msg("Usage: don <obj>")
            return
        armor = self.caller.search(self.args, candidates=self.caller.contents)
        if not armor:
            return
        if not armor.is_typeclass(
            "evennia.contrib.game_systems.turnbattle.tb_equip.TBEArmor", exact=True
        ):
            self.caller.msg("That's not armor!")
            # Remember to update the path to the armor typeclass if you move this module!
            return

        if not self.caller.db.worn_armor:
            self.caller.db.worn_armor = armor
            self.caller.location.msg_contents("%s dons %s." % (self.caller, armor))
        else:
            old_armor = self.caller.db.worn_armor
            self.caller.db.worn_armor = armor
            self.caller.location.msg_contents(
                "%s removes %s and dons %s." % (self.caller, old_armor, armor)
            )


class CmdDoff(Command):
    """
    Stop wearing armor.

    Usage:
      doff

    After using this command, you will stop wearing any
    armor you are currently using and become unarmored.
    You can't use this command in combat.
    """

    key = "doff"
    help_category = "combat"

    rules = COMBAT_RULES

    def func(self):
        """
        This performs the actual command.
        """
        # Can't do this in combat
        if self.rules.is_in_combat(self.caller):
            self.caller.msg("You can't doff armor in a fight!")
            return
        if not self.caller.db.worn_armor:
            self.caller.msg("You aren't wearing any armor!")
        else:
            old_armor = self.caller.db.worn_armor
            self.caller.db.worn_armor = None
            self.caller.location.msg_contents("%s removes %s." % (self.caller, old_armor))


class BattleCmdSet(default_cmds.CharacterCmdSet):
    """
    This command set includes all the commmands used in the battle system.
    """

    key = "DefaultCharacter"

    def at_cmdset_creation(self):
        """
        Populates the cmdset
        """
        self.add(CmdFight())
        self.add(CmdAttack())
        self.add(CmdRest())
        self.add(CmdPass())
        self.add(CmdDisengage())
        self.add(CmdCombatHelp())
        self.add(CmdWield())
        self.add(CmdUnwield())
        self.add(CmdDon())
        self.add(CmdDoff())


"""
----------------------------------------------------------------------------
PROTOTYPES START HERE
----------------------------------------------------------------------------
"""

BASEWEAPON = {"typeclass": "evennia.contrib.game_systems.turnbattle.tb_equip.TBEWeapon"}

BASEARMOR = {"typeclass": "evennia.contrib.game_systems.turnbattle.tb_equip.TBEArmor"}

DAGGER = {
    "prototype": "BASEWEAPON",
    "damage_range": (10, 20),
    "accuracy_bonus": 30,
    "key": "a thin steel dagger",
    "weapon_type_name": "dagger",
}

BROADSWORD = {
    "prototype": "BASEWEAPON",
    "damage_range": (15, 30),
    "accuracy_bonus": 15,
    "key": "an iron broadsword",
    "weapon_type_name": "broadsword",
}

GREATSWORD = {
    "prototype": "BASEWEAPON",
    "damage_range": (20, 40),
    "accuracy_bonus": 0,
    "key": "a rune-etched greatsword",
    "weapon_type_name": "greatsword",
}

LEATHERARMOR = {
    "prototype": "BASEARMOR",
    "damage_reduction": 2,
    "defense_modifier": -2,
    "key": "a suit of leather armor",
}

SCALEMAIL = {
    "prototype": "BASEARMOR",
    "damage_reduction": 4,
    "defense_modifier": -4,
    "key": "a suit of scale mail",
}

PLATEMAIL = {
    "prototype": "BASEARMOR",
    "damage_reduction": 6,
    "defense_modifier": -6,
    "key": "a suit of plate mail",
}



"""
Traits - Whitenoise 2014, Griatch 2020

"""

from .traits import MandatoryTraitKey  # noqa
from .traits import TraitException  # noqa
from .traits import TraitHandler  # noqa
from .traits import TraitProperty  # noqa
from .traits import CounterTrait, GaugeTrait, StaticTrait, Trait  # noqa



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-07-05 17:36


from django.db import migrations

# This migration only made sense earlier in the git history, during
# the player->account transition. Now it will do nothing since players.PlayerDB
# no longer exists.


def forwards(apps, schema_editor):

    try:
        apps.get_model("players", "PlayerDB")
    except LookupError:
        return
    AccountDB = apps.get_model("accounts", "AccountDB")

    Msg = apps.get_model("comms", "Msg")
    for msg in Msg.objects.all():
        for player in msg.db_sender_players.all():
            account = AccountDB.objects.get(id=player.id)
            msg.db_sender_accounts.add(account)
        for player in msg.db_receivers_players.all():
            account = AccountDB.objects.get(id=player.id)
            msg.db_receivers_accounts.add(account)
        for player in msg.db_hide_from_players.all():
            account = AccountDB.objects.get(id=player.id)
            msg.db_hide_from_accounts.add(account)

    ChannelDB = apps.get_model("comms", "ChannelDB")
    for channel in ChannelDB.objects.all():
        for player in channel.db_subscriptions.all():
            account = AccountDB.objects.get(id=player.id)
            channel.db_account_subscriptions.add(player)


class Migration(migrations.Migration):

    dependencies = [("comms", "0013_auto_20170705_1726")]

    operations = [migrations.RunPython(forwards)]



import time
from collections import deque

from django.core.cache import caches
from django.utils.translation import gettext as _

from evennia.utils import logger


class Throttle:
    """
    Keeps a running count of failed actions per IP address.

    Available methods indicate whether or not the number of failures exceeds a
    particular threshold.

    This version of the throttle is usable by both the terminal server as well
    as the web server, imposes limits on memory consumption by using deques
    with length limits instead of open-ended lists, and uses native Django
    caches for automatic key eviction and persistence configurability.
    """

    error_msg = _("Too many failed attempts; you must wait a few minutes before trying again.")

    def __init__(self, **kwargs):
        """
        Allows setting of throttle parameters.

        Keyword Args:
            name (str): Name of this throttle.
            limit (int): Max number of failures before imposing limiter. If `None`,
                the throttle is disabled.
            timeout (int): number of timeout seconds after
                max number of tries has been reached.
            cache_size (int): Max number of attempts to record per IP within a
                rolling window; this is NOT the same as the limit after which
                the throttle is imposed!
        """
        try:
            self.storage = caches["throttle"]
        except Exception:
            logger.log_trace("Throttle: Errors encountered; using default cache.")
            self.storage = caches["default"]

        self.name = kwargs.get("name", "undefined-throttle")
        self.limit = kwargs.get("limit", 5)
        self.cache_size = kwargs.get("cache_size", self.limit)
        self.timeout = kwargs.get("timeout", 5 * 60)

    def get_cache_key(self, *args, **kwargs):
        """
        Creates a 'prefixed' key containing arbitrary terms to prevent key
        collisions in the same namespace.

        """
        return "-".join((self.name, *args))

    def touch(self, key, *args, **kwargs):
        """
        Refreshes the timeout on a given key and ensures it is recorded in the
        key register.

        Args:
            key(str): Key of entry to renew.

        """
        cache_key = self.get_cache_key(key)
        if self.storage.touch(cache_key, self.timeout):
            self.record_key(key)

    def get(self, ip=None):
        """
        Convenience function that returns the storage table, or part of.

        Args:
            ip (str, optional): IP address of requestor

        Returns:
            storage (dict): When no IP is provided, returns a dict of all
                current IPs being tracked and the timestamps of their recent
                failures.
            timestamps (deque): When an IP is provided, returns a deque of
                timestamps of recent failures only for that IP.

        """
        if ip:
            cache_key = self.get_cache_key(str(ip))
            return self.storage.get(cache_key, deque(maxlen=self.cache_size))
        else:
            keys_key = self.get_cache_key("keys")
            keys = self.storage.get_or_set(keys_key, set(), self.timeout)
            data = self.storage.get_many((self.get_cache_key(x) for x in keys))

            found_keys = set(data.keys())
            if len(keys) != len(found_keys):
                self.storage.set(keys_key, found_keys, self.timeout)

            return data

    def update(self, ip, failmsg="Exceeded threshold."):
        """
        Store the time of the latest failure.

        Args:
            ip (str): IP address of requestor
            failmsg (str, optional): Message to display in logs upon activation
                of throttle.

        Returns:
            None

        """
        cache_key = self.get_cache_key(ip)

        # Get current status
        previously_throttled = self.check(ip)

        # Get previous failures, if any
        entries = self.storage.get(cache_key, [])
        entries.append(time.time())

        # Store updated record
        self.storage.set(cache_key, deque(entries, maxlen=self.cache_size), self.timeout)

        # See if this update caused a change in status
        currently_throttled = self.check(ip)

        # If this makes it engage, log a single activation event
        if not previously_throttled and currently_throttled:
            logger.log_sec(
                f"Throttle Activated: {failmsg} (IP: {ip}, "
                f"{self.limit} hits in {self.timeout} seconds.)"
            )

        self.record_ip(ip)

    def remove(self, ip, *args, **kwargs):
        """
        Clears data stored for an IP from the throttle.

        Args:
            ip(str): IP to clear.

        """
        exists = self.get(ip)
        if not exists:
            return False

        cache_key = self.get_cache_key(ip)
        self.storage.delete(cache_key)
        self.unrecord_ip(ip)

        # Return True if NOT exists
        return not bool(self.get(ip))

    def record_ip(self, ip, *args, **kwargs):
        """
        Tracks keys as they are added to the cache (since there is no way to
        get a list of keys after-the-fact).

        Args:
            ip(str): IP being added to cache. This should be the original
                IP, not the cache-prefixed key.

        """
        keys_key = self.get_cache_key("keys")
        keys = self.storage.get(keys_key, set())
        keys.add(ip)
        self.storage.set(keys_key, keys, self.timeout)
        return True

    def unrecord_ip(self, ip, *args, **kwargs):
        """
        Forces removal of a key from the key registry.

        Args:
            ip(str): IP to remove from list of keys.

        """
        keys_key = self.get_cache_key("keys")
        keys = self.storage.get(keys_key, set())
        try:
            keys.remove(ip)
            self.storage.set(keys_key, keys, self.timeout)
            return True
        except KeyError:
            return False

    def check(self, ip):
        """
        This will check the session's address against the
        storage dictionary to check they haven't spammed too many
        fails recently.

        Args:
            ip (str): IP address of requestor

        Returns:
            throttled (bool): True if throttling is active,
                False otherwise.

        """
        if self.limit is None:
            # throttle is disabled
            return False

        now = time.time()
        ip = str(ip)

        cache_key = self.get_cache_key(ip)

        # checking mode
        latest_fails = self.storage.get(cache_key)
        if latest_fails and len(latest_fails) >= self.limit:
            # too many fails recently
            if now - latest_fails[-1] < self.timeout:
                # too soon - timeout in play
                self.touch(cache_key)
                return True
            else:
                # timeout has passed. clear faillist
                self.remove(ip)
                return False
        else:
            return False



"""
This module provides a set of permission lock functions for use
with Evennia's permissions system.

To call these locks, make sure this module is included in the
settings tuple `PERMISSION_FUNC_MODULES` then define a lock on the form
'<access_type>:func(args)' and add it to the object's lockhandler.
Run the `access()` method of the handler to execute the lock check.

Note that `accessing_obj` and `accessed_obj` can be any object type
with a lock variable/field, so be careful to not expect
a certain object type.

"""


from ast import literal_eval

from django.conf import settings

import evennia
from evennia.utils import utils

_PERMISSION_HIERARCHY = [pe.lower() for pe in settings.PERMISSION_HIERARCHY]
# also accept different plural forms
_PERMISSION_HIERARCHY_PLURAL = [
    pe + "s" if not pe.endswith("s") else pe for pe in _PERMISSION_HIERARCHY
]


def _to_account(accessing_obj):
    "Helper function. Makes sure an accessing object is an account object"
    if utils.inherits_from(accessing_obj, "evennia.objects.objects.DefaultObject"):
        # an object. Convert to account.
        accessing_obj = accessing_obj.account
    return accessing_obj


# lock functions


def true(*args, **kwargs):
    """
    Always returns True.

    """
    return True


def all(*args, **kwargs):
    return True


def false(*args, **kwargs):
    """
    Always returns False

    """
    return False


def none(*args, **kwargs):
    return False


def superuser(*args, **kwargs):
    return False


def self(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Check if accessing_obj is the same as accessed_obj

    Usage:
       self()

    This can be used to lock specifically only to
    the same object that the lock is defined on.
    """
    return accessing_obj == accessed_obj


def perm(accessing_obj, accessed_obj, *args, **kwargs):
    """
    The basic permission-checker. Ignores case.

    Usage:
       perm(<permission>)

    where <permission> is the permission accessing_obj must
    have in order to pass the lock.

    If the given permission is part of settings.PERMISSION_HIERARCHY,
    permission is also granted to all ranks higher up in the hierarchy.

    If accessing_object is an Object controlled by an Account, the
    permissions of the Account is used unless the Attribute _quell
    is set to True on the Object. In this case however, the
    LOWEST hieararcy-permission of the Account/Object-pair will be used
    (this is order to avoid Accounts potentially escalating their own permissions
    by use of a higher-level Object)

    For non-hierarchical permissions, a puppeted object's account is checked first,
    followed by the puppet (unless quelled, when only puppet's access is checked).

    """
    # this allows the perm_above lockfunc to make use of this function too
    try:
        permission = args[0].lower()
        perms_object = accessing_obj.permissions.all()
    except (AttributeError, IndexError):
        return False

    gtmode = kwargs.pop("_greater_than", False)
    is_quell = False

    account = (
        utils.inherits_from(accessing_obj, "evennia.objects.objects.DefaultObject")
        and accessing_obj.account
    )
    # check object perms (note that accessing_obj could be an Account too)
    perms_account = []
    if account:
        perms_account = account.permissions.all()
        is_quell = account.attributes.get("_quell")

    # Check hirarchy matches; handle both singular/plural forms in hierarchy
    hpos_target = None
    if permission in _PERMISSION_HIERARCHY:
        hpos_target = _PERMISSION_HIERARCHY.index(permission)
    elif permission.endswith("s") and permission[:-1] in _PERMISSION_HIERARCHY:
        hpos_target = _PERMISSION_HIERARCHY.index(permission[:-1])
    if hpos_target is not None:
        # hieratchy match
        hpos_account = -1
        hpos_object = -1

        if account:
            # we have an account puppeting this object. We must check what perms it has
            perms_account_single = [p[:-1] if p.endswith("s") else p for p in perms_account]
            hpos_account = [
                hpos
                for hpos, hperm in enumerate(_PERMISSION_HIERARCHY)
                if hperm in perms_account_single
            ]
            hpos_account = hpos_account[-1] if hpos_account else -1

        if not account or is_quell:
            # only get the object-level perms if there is no account or quelling
            perms_object_single = [p[:-1] if p.endswith("s") else p for p in perms_object]
            hpos_object = [
                hpos
                for hpos, hperm in enumerate(_PERMISSION_HIERARCHY)
                if hperm in perms_object_single
            ]
            hpos_object = hpos_object[-1] if hpos_object else -1

        if account and is_quell:
            # quell mode: use smallest perm from account and object
            if gtmode:
                return hpos_target < min(hpos_account, hpos_object)
            else:
                return hpos_target <= min(hpos_account, hpos_object)
        elif account:
            # use account perm
            if gtmode:
                return hpos_target < hpos_account
            else:
                return hpos_target <= hpos_account
        else:
            # use object perm
            if gtmode:
                return hpos_target < hpos_object
            else:
                return hpos_target <= hpos_object
    else:
        # no hierarchy match - check direct matches
        if account:
            # account exists
            if is_quell and permission in perms_object:
                # if quelled, first check object
                return True
            elif permission in perms_account:
                # unquelled - check account
                return True
            else:
                # no account-pass, check object pass
                return permission in perms_object

        elif permission in perms_object:
            return True

    return False


def perm_above(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Only allow objects with a permission *higher* in the permission
    hierarchy than the one given. If there is no such higher rank,
    it's assumed we refer to superuser. If no hierarchy is defined,
    this function has no meaning and returns False.
    """
    kwargs["_greater_than"] = True
    return perm(accessing_obj, accessed_obj, *args, **kwargs)


def pperm(accessing_obj, accessed_obj, *args, **kwargs):
    """
    The basic permission-checker only for Account objects. Ignores case.

    Usage:
       pperm(<permission>)

    where <permission> is the permission accessing_obj must
    have in order to pass the lock. If the given permission
    is part of _PERMISSION_HIERARCHY, permission is also granted
    to all ranks higher up in the hierarchy.
    """
    return perm(_to_account(accessing_obj), accessed_obj, *args, **kwargs)


def pperm_above(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Only allow Account objects with a permission *higher* in the permission
    hierarchy than the one given. If there is no such higher rank,
    it's assumed we refer to superuser. If no hierarchy is defined,
    this function has no meaning and returns False.
    """
    return perm_above(_to_account(accessing_obj), accessed_obj, *args, **kwargs)


def dbref(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
      dbref(3)

    This lock type checks if the checking object
    has a particular dbref. Note that this only
    works for checking objects that are stored
    in the database (e.g. not for commands)
    """
    if not args:
        return False
    try:
        dbr = int(args[0].strip().strip("#"))
    except ValueError:
        return False
    if hasattr(accessing_obj, "dbid"):
        return dbr == accessing_obj.dbid
    return False


def pdbref(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Same as dbref, but making sure accessing_obj is an account.
    """
    return dbref(_to_account(accessing_obj), accessed_obj, *args, **kwargs)


def id(accessing_obj, accessed_obj, *args, **kwargs):
    "Alias to dbref"
    return dbref(accessing_obj, accessed_obj, *args, **kwargs)


def pid(accessing_obj, accessed_obj, *args, **kwargs):
    "Alias to dbref, for Accounts"
    return dbref(_to_account(accessing_obj), accessed_obj, *args, **kwargs)


# this is more efficient than multiple if ... elif statments
CF_MAPPING = {
    "eq": lambda val1, val2: val1 == val2 or str(val1) == str(val2) or float(val1) == float(val2),
    "gt": lambda val1, val2: float(val1) > float(val2),
    "lt": lambda val1, val2: float(val1) < float(val2),
    "ge": lambda val1, val2: float(val1) >= float(val2),
    "le": lambda val1, val2: float(val1) <= float(val2),
    "ne": lambda val1, val2: float(val1) != float(val2),
    "default": lambda val1, val2: False,
}


def attr(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
      attr(attrname)
      attr(attrname, value)
      attr(attrname, value, compare=type)

    where compare's type is one of (eq,gt,lt,ge,le,ne) and signifies
    how the value should be compared with one on accessing_obj (so
    compare=gt means the accessing_obj must have a value greater than
    the one given).

    Searches attributes *and* properties stored on the accessing_obj.
    if accessing_obj has a property "obj", then this is used as
    accessing_obj (this makes this usable for Commands too)

    The first form works like a flag - if the attribute/property
    exists on the object, the value is checked for True/False. The
    second form also requires that the value of the attribute/property
    matches. Note that all retrieved values will be converted to
    strings before doing the comparison.
    """
    # deal with arguments
    if not args:
        return False
    attrname = args[0].strip()
    value = None
    if len(args) > 1:
        value = args[1].strip()
    compare = "eq"
    if kwargs:
        compare = kwargs.get("compare", "eq")

    def valcompare(val1, val2, typ="eq"):
        "compare based on type"
        try:
            return CF_MAPPING.get(typ, CF_MAPPING["default"])(val1, val2)
        except Exception:
            # this might happen if we try to compare two things that
            # cannot be compared
            return False

    if hasattr(accessing_obj, "obj"):
        # NOTE: this is relevant for Commands. It may clash with scripts
        # (they have Attributes and .obj) , but are scripts really
        # used so that one ever wants to check the property on the
        # Script rather than on its owner?
        accessing_obj = accessing_obj.obj

    # first, look for normal properties on the object trying to gain access
    if hasattr(accessing_obj, attrname):
        if value:
            return valcompare(str(getattr(accessing_obj, attrname)), value, compare)
        # will return Fail on False value etc
        return bool(getattr(accessing_obj, attrname))
    # check attributes, if they exist
    if hasattr(accessing_obj, "attributes") and accessing_obj.attributes.has(attrname):
        if value:
            return hasattr(accessing_obj, "attributes") and valcompare(
                accessing_obj.attributes.get(attrname), value, compare
            )
        # fails on False/None values
        return bool(accessing_obj.attributes.get(attrname))
    return False


def objattr(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
      objattr(attrname)
      objattr(attrname, value)
      objattr(attrname, value, compare=type)

    Works like attr, except it looks for an attribute on
    accessed_obj instead.

    """
    return attr(accessed_obj, accessed_obj, *args, **kwargs)


def locattr(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
      locattr(attrname)
      locattr(attrname, value)
      locattr(attrname, value, compare=type)

    Works like attr, except it looks for an attribute on
    accessing_obj.location, if such an entity exists.

    if accessing_obj has a property ".obj" (such as is the case for a
    Command), then accessing_obj.obj.location is used instead.

    """
    if hasattr(accessing_obj, "obj"):
        accessing_obj = accessing_obj.obj
    if hasattr(accessing_obj, "location"):
        return attr(accessing_obj.location, accessed_obj, *args, **kwargs)
    return False


def objlocattr(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
      locattr(attrname)
      locattr(attrname, value)
      locattr(attrname, value, compare=type)

    Works like attr, except it looks for an attribute on
    accessed_obj.location, if such an entity exists.

    if accessed_obj has a property ".obj" (such as is the case for a
    Command), then accessing_obj.obj.location is used instead.

    """
    if hasattr(accessed_obj, "obj"):
        accessed_obj = accessed_obj.obj
    if hasattr(accessed_obj, "location"):
        return attr(accessed_obj.location, accessed_obj, *args, **kwargs)
    return False


def attr_eq(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
       attr_gt(attrname, 54)
    """
    return attr(accessing_obj, accessed_obj, *args, **kwargs)


def attr_gt(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
       attr_gt(attrname, 54)

    Only true if access_obj's attribute > the value given.
    """
    return attr(accessing_obj, accessed_obj, *args, **{"compare": "gt"})


def attr_ge(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
       attr_gt(attrname, 54)

    Only true if access_obj's attribute >= the value given.
    """
    return attr(accessing_obj, accessed_obj, *args, **{"compare": "ge"})


def attr_lt(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
       attr_gt(attrname, 54)

    Only true if access_obj's attribute < the value given.
    """
    return attr(accessing_obj, accessed_obj, *args, **{"compare": "lt"})


def attr_le(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
       attr_gt(attrname, 54)

    Only true if access_obj's attribute <= the value given.
    """
    return attr(accessing_obj, accessed_obj, *args, **{"compare": "le"})


def attr_ne(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
       attr_gt(attrname, 54)

    Only true if access_obj's attribute != the value given.
    """
    return attr(accessing_obj, accessed_obj, *args, **{"compare": "ne"})


def tag(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
        tag(tagkey)
        tag(tagkey, category)

    Only true if accessing_obj has the specified tag and optional
    category.
    If accessing_obj has the ".obj" property (such as is the case for
    a command), then accessing_obj.obj is used instead.

    """
    if hasattr(accessing_obj, "obj"):
        accessing_obj = accessing_obj.obj
    tagkey = args[0] if args else None
    category = args[1] if len(args) > 1 else None
    return bool(accessing_obj.tags.get(tagkey, category=category))


def objtag(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
        objtag(tagkey)
        objtag(tagkey, category):

    Only true if `accessed_obj` has the given tag and optional category.

    """
    return tag(accessed_obj, None, *args, **kwargs)


def objloctag(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
        objloctag(tagkey)
        objloctag(tagkey, category):

    Only true if `accessed_obj.location` has the given tag and optional category.
    If obj has no location, this lockfunc fails.

    """
    try:
        return tag(accessed_obj.location, None, *args, **kwargs)
    except AttributeError:
        return False


def is_ooc(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
        is_ooc()

    This is normally used to lock a Command, so it can be used
    only when out of character. When not logged in at all, this
    function will still return True.
    """
    obj = accessed_obj.obj if hasattr(accessed_obj, "obj") else accessed_obj
    account = obj.account if utils.inherits_from(obj, evennia.DefaultObject) else obj
    if not account:
        return True
    try:
        session = accessed_obj.session
    except AttributeError:
        # note-this doesn't work well
        # for high multisession mode. We may need
        # to change to sessiondb to resolve this
        sessions = session = account.sessions.get()
        session = sessions[0] if sessions else None
    if not session:
        # this suggests we are not even logged in; treat as ooc.
        return True
    try:
        return not account.get_puppet(session)
    except TypeError:
        return not session.get_puppet()


def objtag(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
        objtag(tagkey)
        objtag(tagkey, category)

    Only true if accessed_obj has the specified tag and optional
    category.
    """
    tagkey = args[0] if args else None
    category = args[1] if len(args) > 1 else None
    return bool(accessed_obj.tags.get(tagkey, category=category))


def inside(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
       inside()

    True if accessing_obj is 'inside' accessing_obj. Note that this only checks
    one level down. So if if the lock is on a room, you will pass but not your
    inventory (since their location is you, not the locked object).  If you
    want also nested objects to pass the lock, use the `insiderecursive`
    lockfunc.
    """
    if hasattr(accessed_obj, "obj"):
        accessed_obj = accessed_obj.obj
    return accessing_obj.location == accessed_obj


def inside_rec(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
        inside_rec()

    True if accessing_obj is inside the accessed obj, at up to 10 levels
    of recursion (so if this lock is on a room, then an object inside a box
    in your inventory will also pass the lock).
    """

    def _recursive_inside(obj, accessed_obj, lvl=1):
        if obj.location:
            if obj.location == accessed_obj:
                return True
            elif lvl >= 10:
                # avoid infinite recursions
                return False
            else:
                return _recursive_inside(obj.location, accessed_obj, lvl + 1)
        return False

    return _recursive_inside(accessing_obj, accessed_obj)


def holds(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Usage:
      holds()            checks if accessed_obj or accessed_obj.obj
                         is held by accessing_obj
      holds(key/dbref)   checks if accessing_obj holds an object
                          with given key/dbref
      holds(attrname, value)   checks if accessing_obj holds an
                               object with the given attrname and value

    This is passed if accessed_obj is carried by accessing_obj (that is,
    accessed_obj.location == accessing_obj), or if accessing_obj itself holds
    an object matching the given key.
    """
    try:
        # commands and scripts don't have contents, so we are usually looking
        # for the contents of their .obj property instead (i.e. the object the
        # command/script is attached to).
        contents = accessing_obj.contents
    except AttributeError:
        try:
            contents = accessing_obj.obj.contents
        except AttributeError:
            return False

    def check_holds(objid):
        # helper function. Compares both dbrefs and keys/aliases.
        objid = str(objid)
        dbref = utils.dbref(objid, reqhash=False)
        if dbref and any((True for obj in contents if obj.dbid == dbref)):
            return True
        objid = objid.lower()
        return any(
            (
                True
                for obj in contents
                if obj.key.lower() == objid or objid in [al.lower() for al in obj.aliases.all()]
            )
        )

    if not args:
        # holds() - check if accessed_obj or accessed_obj.obj is held by accessing_obj
        try:
            if check_holds(accessed_obj.dbid):
                return True
        except Exception:
            # we need to catch any trouble here
            pass
        return hasattr(accessed_obj, "obj") and check_holds(accessed_obj.obj.dbid)
    if len(args) == 1:
        # command is holds(dbref/key) - check if given objname/dbref is held by accessing_ob
        return check_holds(args[0])
    elif len(args) > 1:
        # command is holds(attrname, value) check if any held object has the given attribute and value
        for obj in contents:
            if obj.attributes.get(args[0]) == args[1]:
                return True
    return False


def has_account(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Only returns true if accessing_obj has_account is true, that is,
    this is an account-controlled object. It fails on actual accounts!

    This is a useful lock for traverse-locking Exits to restrain NPC
    mobiles from moving outside their areas.
    """
    return utils.inherits_from(accessing_obj, evennia.DefaultObject) and accessing_obj.has_account


def serversetting(accessing_obj, accessed_obj, *args, **kwargs):
    """
    Only returns true if the Evennia settings exists, alternatively has
    a certain value.

    Usage:
      serversetting(IRC_ENABLED)
      serversetting(BASE_SCRIPT_PATH, ['types'])

    A given True/False or integers will be converted properly. Note that
    everything will enter this function as strings, so they have to be
    unpacked to their real value. We only support basic properties.
    """
    if not args or not args[0]:
        return False
    if len(args) < 2:
        setting = args[0]
        val = "True"
    else:
        setting, val = args[0], args[1]
    # convert
    try:
        val = literal_eval(val)
    except Exception:
        # we swallow errors here, lockfuncs has noone to report to
        return False

    if setting in settings._wrapped.__dict__:
        return settings._wrapped.__dict__[setting] == val
    return False



"""
Test the EvAdventure commands.

"""

from unittest.mock import call, patch

from anything import Something

from evennia.utils.create import create_object
from evennia.utils.test_resources import BaseEvenniaCommandTest

from .. import commands
from ..characters import EvAdventureCharacter
from ..npcs import EvAdventureMob, EvAdventureShopKeeper
from .mixins import EvAdventureMixin


class TestEvAdventureCommands(EvAdventureMixin, BaseEvenniaCommandTest):
    def setUp(self):
        super().setUp()
        # needed for the .call mechanism
        self.char1 = self.character

    def test_inventory(self):
        self.call(
            commands.CmdInventory(),
            "inventory",
            """
You are fighting with your bare fists and have no shield.
You wear no armor and no helmet.
Backpack is empty.
You use 0/11 equipment slots.
""".strip(),
        )

    def test_wield_or_wear(self):
        self.char1.equipment.add(self.helmet)
        self.char1.equipment.add(self.weapon)
        self.shield.location = self.location

        self.call(commands.CmdWieldOrWear(), "shield", "Could not find 'shield'")
        self.call(commands.CmdWieldOrWear(), "helmet", "You put helmet on your head.")
        self.call(
            commands.CmdWieldOrWear(),
            "weapon",
            "You hold weapon in your strongest hand, ready for action.",
        )
        self.call(commands.CmdWieldOrWear(), "helmet", "You are already using helmet.")

    def test_remove(self):
        self.char1.equipment.add(self.helmet)
        self.call(commands.CmdWieldOrWear(), "helmet", "You put helmet on your head.")

        self.call(commands.CmdRemove(), "helmet", "You stash helmet in your backpack.")

    def test_give__coins(self):
        recipient = create_object(EvAdventureCharacter, key="Friend", location=self.location)
        recipient.coins = 0
        self.char1.coins = 100

        self.call(commands.CmdGive(), "40 coins to friend", "You give Friend 40 coins.")
        self.assertEqual(self.char1.coins, 60)
        self.assertEqual(recipient.coins, 40)

        self.call(commands.CmdGive(), "10 to friend", "You give Friend 10 coins.")
        self.assertEqual(self.char1.coins, 50)
        self.assertEqual(recipient.coins, 50)

        self.call(commands.CmdGive(), "60 to friend", "You only have 50 coins to give.")

        recipient.delete()

    @patch("evennia.contrib.tutorials.evadventure.commands.EvMenu")
    def test_give__item(self, mock_EvMenu):
        self.char1.equipment.add(self.helmet)
        recipient = create_object(EvAdventureCharacter, key="Friend", location=self.location)

        self.call(commands.CmdGive(), "helmet to friend", "")

        mock_EvMenu.assert_has_calls(
            (
                call(
                    recipient,
                    {"node_receive": Something, "node_end": Something},
                    item=self.helmet,
                    giver=self.char1,
                ),
                call(
                    self.char1,
                    {"node_give": Something, "node_end": Something},
                    item=self.helmet,
                    receiver=recipient,
                ),
            )
        )

        recipient.delete()

    @patch("evennia.contrib.tutorials.evadventure.npcs.EvMenu")
    def test_talk(self, mock_EvMenu):
        npc = create_object(EvAdventureShopKeeper, key="shopkeep", location=self.location)

        npc.menudata = {"foo": None, "bar": None}

        self.call(commands.CmdTalk(), "shopkeep", "")

        mock_EvMenu.assert_called_with(
            self.char1,
            {"foo": None, "bar": None},
            startnode="node_start",
            session=None,
            npc=npc,
        )



from . import exceptions

COMPONENT_LISTING = {}


def get_component_class(name):
    """
    Retrieves a component from the listing using a name
    Args:
        name (str): The unique name of the component
    """
    component_class = COMPONENT_LISTING.get(name)
    if component_class is None:
        message = (
            f"Component with name {name} has not been found. "
            "Make sure it has been imported before being used."
        )
        raise exceptions.ComponentDoesNotExist(message)

    return component_class
